import zigpy_znp.types as t


class AttributeValue(t.FixedList, item_type=t.uint8_t, length=16):
    pass


class MAC(t.CommandsBase, subsystem=t.Subsystem.MAC):
    # MAC Reset command to reset MAC state machine
    ResetReq = t.CommandDef(
        t.CommandType.SREQ,
        0x01,
        req_schema=(
            t.Param(
                "SetDefault",
                t.Bool,
                "TRUE - Set the MAC pib values to default values",
            ),
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # initialize the MAC
    Init = t.CommandDef(
        t.CommandType.SREQ, 0x02, req_schema=(), rsp_schema=t.STATUS_SCHEMA
    )

    # start the MAC as a coordinator or end device
    StartReq = t.CommandDef(
        t.CommandType.SREQ,
        0x03,
        req_schema=(
            t.Param(
                "StartTime",
                t.uint32_t,
                (
                    "The time to begin transmitting beacons relative to "
                    "the received beacon"
                ),
            ),
            t.Param(
                "PanId",
                t.PanId,
                (
                    "The PAN Id to use. This parameter is ignored if Pan "
                    "Coordinator is FALSE"
                ),
            ),
            t.Param("LogicalChannel", t.uint8_t, "The logical channel to use"),
            t.Param("ChannelPage", t.uint8_t, "The channel page to use"),
            t.Param(
                "BeaconOrder",
                t.uint8_t,
                "The exponent used to calculate the beacon interval",
            ),
            t.Param(
                "SuperFrameOrder",
                t.uint8_t,
                "The exponent used to calculate the superframe duration",
            ),
            t.Param(
                "PanCoordinator",
                t.Bool,
                "Set to TRUE to start a network as PAN coordinator",
            ),
            t.Param(
                "BatteryLifeExt",
                t.uint8_t,
                "full backoff periods following the interframe spacing",
            ),
            t.Param("CoordRealignment", t.uint8_t, "Coordinator realignment"),
            t.Param("RealignKeySource", t.KeySource, "Key Source of this data frame"),
            # ToDo: Enum for for RealignSecurityLevel
            t.Param(
                "RealignSecurityLevel",
                t.uint8_t,
                "Security level of this data frame",
            ),
            # ToDo: Make this an enum
            t.Param("RealignKeyIdMode", t.uint8_t, "Key Id Mode of this frame"),
            t.Param("RealignKeyIndex", t.uint8_t, "Key index of this frame"),
            t.Param("BeaconKeySource", t.KeySource, "Key source of this data frame"),
            # ToDo: Make this an enum
            t.Param(
                "BeaconSecurityLevel",
                t.uint8_t,
                "Security Level of this data frame",
            ),
            t.Param("BeaconKeyIdMode", t.uint8_t, "Key Id Mode of this data frame"),
            t.Param("BeaconKeyIndex", t.uint8_t, "Key index of this data frame"),
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # request synchronization to the current network beacon
    SyncReq = t.CommandDef(
        t.CommandType.SREQ,
        0x04,
        req_schema=(
            t.Param("LogicalChannel", t.uint8_t, "The logical channel to use"),
            t.Param("ChannelPage", t.uint8_t, "The channel page to use"),
            t.Param(
                "TrackBeacon",
                t.Bool,
                (
                    "Set to TRUE to continue tracking beacons after synchronizing "
                    "with the first beacon. Set to FALSE to only synchronize with "
                    "the first beacon"
                ),
            ),
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # send (on behalf of the next higher layer) MAC Data Frame packet
    DataReq = t.CommandDef(
        t.CommandType.SREQ,
        0x05,
        req_schema=(
            t.Param(
                "DstAddrModeAddress",
                t.AddrModeAddress,
                "Destination address mode and address",
            ),
            t.Param("DstPanId", t.PanId, "The PAN Id of destination"),
            t.Param("SrcAddrMode", t.AddrMode, "Format of the source address"),
            t.Param("Handle", t.uint8_t, "Handle of the packet"),
            # ToDo: Make this a proper Flags Enum
            t.Param("TxOption", t.uint8_t, "Transmitting options"),
            t.Param(
                "LogicalChannel",
                t.uint8_t,
                "Channel that data frame will be transmitted",
            ),
            t.Param("Power", t.uint8_t, "Power level to use for transmission"),
            t.Param("KeySource", t.KeySource, "Key Source of this data frame"),
            # ToDo: Make this an enum
            t.Param("SecurityLevel", t.uint8_t, "Security level of this data frame"),
            # ToDo: Make this an enum
            t.Param("KeyIdMode", t.uint8_t, "Key Id Mode of this frame"),
            t.Param("KeyIndex", t.uint8_t, "Key index of this frame"),
            t.Param("MSDU", t.ShortBytes, "Actual data that will be sent"),
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # request (on behalf of the next higher layer) an association with a coordinator
    AssociateReq = t.CommandDef(
        t.CommandType.SREQ,
        0x06,
        req_schema=(
            t.Param("LogicalChannel", t.uint8_t, "The logical channel to use"),
            t.Param("ChannelPage", t.uint8_t, "The channel page to use"),
            t.Param(
                "CoordAddrModeAddress",
                t.AddrModeAddress,
                "Coordinator address mode and address",
            ),
            t.Param("CoordPanId", t.PanId, "The PAN Id of the coordinator"),
            # ToDo: make this a bitflag enum
            t.Param("CapabilityInformation", t.uint8_t, "BitFlag Coordinator"),
            t.Param("KeySource", t.KeySource, "Key Source of this data frame"),
            # ToDo: Make this an enum
            t.Param("SecurityLevel", t.uint8_t, "Security level of this data frame"),
            # ToDo: Make this an enum
            t.Param("KeyIdMode", t.uint8_t, "Key Id Mode of this frame"),
            t.Param("KeyIndex", t.uint8_t, "Key index of this frame"),
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # This command is sent by the host to response to the MAC_ASSOCIATE_IND
    AssociateRsp = t.CommandDef(
        t.CommandType.SREQ,
        0x50,
        req_schema=(
            t.Param(
                "IEEE",
                t.EUI64,
                "Extended address of the device requesting association",
            ),
            t.Param("NWK", t.NWK, "Short address of the associated device"),
            # ToDo: make this an enum
            t.Param("AssocStatus", t.uint8_t, "Status of the associaiton"),
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    #  request (on behalf of the next higher layer) a disassociation of the device
    #  from the coordinator
    DisAssociateReq = t.CommandDef(
        t.CommandType.SREQ,
        0x07,
        req_schema=(
            t.Param(
                "DeviceAddrModeAddress",
                t.AddrModeAddress,
                "Device address mode and address",
            ),
            t.Param("DevicePanId", t.PanId, "Device's PAN Id"),
            # ToDo: Make this an enum
            t.Param("DisassociateReason", t.uint8_t, "Reason for disassociation"),
            t.Param("TxIndirect", t.Bool, "Indirect Transmission"),
            t.Param("KeySource", t.KeySource, "Key Source of this data frame"),
            # ToDo: Make this an enum
            t.Param("SecurityLevel", t.uint8_t, "Security level of this data frame"),
            # ToDo: Make this an enum
            t.Param("KeyIdMode", t.uint8_t, "Key Id Mode of this frame"),
            t.Param("KeyIndex", t.uint8_t, "Key index of this frame"),
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # read (on behalf of the next higher layer) a MAC PIB attribute
    GetReq = t.CommandDef(
        t.CommandType.SREQ,
        0x08,
        req_schema=(
            # ToDo: Make this an enum
            t.Param("Attribute", t.uint8_t, "MAC PIB Attribute to get"),
        ),
        rsp_schema=(
            t.Param("Status", t.Status, "Status is either Success (0) or Failure (1)"),
            t.Param("Value", AttributeValue, "Value of the attribute"),
        ),
    )

    # request the device to write a MAC PIB value
    SetReq = t.CommandDef(
        t.CommandType.SREQ,
        0x09,
        req_schema=(
            # ToDo: Make this an enum
            t.Param("Attribute", t.uint8_t, "MAC PIB Attribute to set"),
            t.Param("Value", AttributeValue, "Value of the attribute"),
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # send a request to the device to perform a network scan
    ScanReq = t.CommandDef(
        t.CommandType.SREQ,
        0x0C,
        req_schema=(
            t.Param(
                "ScanChannels",
                t.Channels,
                "Bitmask of channels to scan when starting the device",
            ),
            # ToDo: Make this an enum
            t.Param("ScanType", t.uint8_t, "Specifies the scan type"),
            t.Param(
                "ScanDuration", t.uint8_t, "The exponent used in the scan duration"
            ),
            t.Param("ChannelPage", t.uint8_t, "The channel page to use"),
            t.Param("MaxResults", t.uint8_t, "Max results (UNDOCUMENTED)"),
            t.Param("KeySource", t.KeySource, "Key Source of this data frame"),
            # ToDo: Make this an enum
            t.Param("SecurityLevel", t.uint8_t, "Security level of this data frame"),
            # ToDo: Make this an enum
            t.Param("KeyIdMode", t.uint8_t, "Key Id Mode of this frame"),
            t.Param("KeyIndex", t.uint8_t, "Key index of this frame"),
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # This command is sent by the host to response to the ORPHAN_IND
    OrphanRsp = t.CommandDef(
        t.CommandType.SREQ,
        0x51,
        req_schema=(
            t.Param(
                "IEEE",
                t.EUI64,
                "Extended address of the device requesting association",
            ),
            t.Param("NWK", t.NWK, "Short address of the associated device"),
            t.Param(
                "AssociatedMember",
                t.Bool,
                "True is the orphan is a associated member",
            ),
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # send a MAC data request poll
    PollReq = t.CommandDef(
        t.CommandType.SREQ,
        0x0D,
        req_schema=(
            t.Param(
                "CoordAddrModeAddress",
                t.AddrModeAddress,
                "Coordinator address mode and address",
            ),
            t.Param("CoordPanId", t.PanId, "The PAN Id of the coordinator"),
            t.Param("KeySource", t.KeySource, "Key Source of this data frame"),
            # ToDo: Make this an enum
            t.Param("SecurityLevel", t.uint8_t, "Security level of this data frame"),
            # ToDo: Make this an enum
            t.Param("KeyIdMode", t.uint8_t, "Key Id Mode of this frame"),
            t.Param("KeyIndex", t.uint8_t, "Key index of this frame"),
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # send a request to the device to purge a data frame
    PurgeReq = t.CommandDef(
        t.CommandType.SREQ,
        0x0E,
        req_schema=(t.Param("MsduHandle", t.uint8_t, "MSDU handle"),),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # send a request to the device to set Rx gain
    SetRxGainReq = t.CommandDef(
        t.CommandType.SREQ,
        0x0F,
        req_schema=(t.Param("Mode", t.Bool, "PA/PNA mode - True/False"),),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # MAC Callbacks

    # send (on behalf of the next higher layer) an indication of the synchronization
    # loss
    SyncLossInd = t.CommandDef(
        t.CommandType.AREQ,
        0x80,
        rsp_schema=(
            t.Param("Status", t.Status, "Status is either Success (0) or Failure (1)"),
            t.Param(
                "PanId",
                t.PanId,
                "The PAN Id to use. This parameter is ignored if Pan",
            ),
            t.Param("LogicalChannel", t.uint8_t, "The logical channel to use"),
            t.Param("ChannelPage", t.uint8_t, "The channel page to use"),
            t.Param("KeySource", t.KeySource, "Key Source of this data frame"),
            # ToDo: Make this an enum
            t.Param("SecurityLevel", t.uint8_t, "Security level of this data frame"),
            # ToDo: Make this an enum
            t.Param("KeyIdMode", t.uint8_t, "Key Id Mode of this frame"),
            t.Param("KeyIndex", t.uint8_t, "Key index of this frame"),
        ),
    )

    # send (on behalf of the next higher layer) an association indication message
    AssociateInd = t.CommandDef(
        t.CommandType.AREQ,
        0x81,
        rsp_schema=(
            t.Param("IEEE", t.EUI64, "Extended address of the device"),
            t.Param("Capabilities", t.uint8_t, "Operating capabilities of the device"),
            t.Param("KeySource", t.KeySource, "Key Source of this data frame"),
            # ToDo: Make this an enum
            t.Param("SecurityLevel", t.uint8_t, "Security level of this data frame"),
            # ToDo: Make this an enum
            t.Param("KeyIdMode", t.uint8_t, "Key Id Mode of this frame"),
            t.Param("KeyIndex", t.uint8_t, "Key index of this frame"),
        ),
    )

    # send (on behalf of the next higher layer) an association confirmation message
    AssociateCnf = t.CommandDef(
        t.CommandType.AREQ,
        0x82,
        rsp_schema=(
            t.Param("Status", t.Status, "Status is either Success (0) or Failure (1)"),
            t.Param("NWK", t.NWK, "Short address of the device"),
            t.Param("KeySource", t.KeySource, "Key Source of this data frame"),
            # ToDo: Make this an enum
            t.Param("SecurityLevel", t.uint8_t, "Security level of this data frame"),
            # ToDo: Make this an enum
            t.Param("KeyIdMode", t.uint8_t, "Key Id Mode of this frame"),
            t.Param("KeyIndex", t.uint8_t, "Key index of this frame"),
        ),
    )

    # send (on behalf of the next higher layer) a MAC beacon notify indication
    BeaconNotifyInd = t.CommandDef(
        t.CommandType.AREQ,
        0x83,
        rsp_schema=(
            t.Param("BSN", t.uint8_t, "BSN"),
            t.Param("TimeStamp", t.uint32_t, "The timestamp of the message"),
            t.Param(
                "CoordinatorExtendedAddress",
                t.AddrModeAddress,
                "Extended address of coordinator",
            ),
            t.Param(
                "PanId",
                t.PanId,
                "The PAN Id to use. This parameter is ignored if Pan",
            ),
            t.Param("Superframe", t.uint16_t, "Superframe specification"),
            t.Param("LogicalChannel", t.uint8_t, "The logical channel to use"),
            t.Param("GTSPermit", t.Bool, "True/False - Permit/Not permit GTS"),
            t.Param("LQI", t.uint8_t, "Link quality of the message"),
            t.Param("SecurityFailure", t.uint8_t, "Security failure???"),
            t.Param("KeySource", t.KeySource, "Key Source of this data frame"),
            # ToDo: Make this an enum
            t.Param("SecurityLevel", t.uint8_t, "Security level of this data frame"),
            # ToDo: Make this an enum
            t.Param("KeyIdMode", t.uint8_t, "Key Id Mode of this frame"),
            t.Param("KeyIndex", t.uint8_t, "Key index of this frame"),
            t.Param("PendingAddrSpec", t.uint8_t, "Pending address spec"),
            t.Param(
                "AddressList",
                t.uint8_t,
                "List of address associate with the device",
            ),
            t.Param("NSDU", t.ShortBytes, "Beacon payload"),
        ),
    )

    # send (on behalf of the next higher layer) a MAC data confirmation
    DataCnf = t.CommandDef(
        t.CommandType.AREQ,
        0x84,
        rsp_schema=(
            t.Param("Status", t.Status, "Status is either Success (0) or Failure (1)"),
            t.Param("Handle", t.uint8_t, "Handle of the message"),
            t.Param("TimeStamp", t.uint32_t, "The timestamp of the message"),
            t.Param("TimeStamp2", t.uint16_t, "16 bit timestamp of the message"),
        ),
    )

    # send (on behalf of the next higher layer) a MAC data indication
    DataInd = t.CommandDef(
        t.CommandType.AREQ,
        0x85,
        rsp_schema=(
            t.Param("SrcAddrModeAddr", t.AddrModeAddress, "Source address"),
            t.Param("DstAddrModeAddr", t.AddrModeAddress, "Destination address"),
            t.Param("TimeStamp", t.uint32_t, "The timestamp of the message"),
            t.Param("TimeStamp2", t.uint16_t, "16 bit timestamp of the message"),
            t.Param("SrcPanId", t.PanId, "PAN Id ofo the source address"),
            t.Param("DstPanId", t.PanId, "PAN Id ofo the destination address"),
            t.Param("LQI", t.uint8_t, "Link quality of the message"),
            t.Param("Correlation", t.uint8_t, "Correlation"),
            t.Param("RSSI", t.int8s, "RSSI"),
            t.Param("DSN", t.uint8_t, "DSN"),
            t.Param("KeySource", t.KeySource, "Key Source of this data frame"),
            # ToDo: Make this an enum
            t.Param("SecurityLevel", t.uint8_t, "Security level of this data frame"),
            # ToDo: Make this an enum
            t.Param("KeyIdMode", t.uint8_t, "Key Id Mode of this frame"),
            t.Param("KeyIndex", t.uint8_t, "Key index of this frame"),
            t.Param("Data", t.ShortBytes, "Actual data that will be sent"),
        ),
    )

    # send (on behalf of the next higher layer) a MAC disassociation indication
    DisassociateReq = t.CommandDef(
        t.CommandType.AREQ,
        0x86,
        rsp_schema=(
            t.Param("IEEE", t.EUI64, "EUI64 address of the device leaving the network"),
            # ToDo: Make this an enum
            t.Param("DisassociateReason", t.uint8_t, "Reason for disassociation"),
            t.Param("KeySource", t.KeySource, "Key Source of this data frame"),
            # ToDo: Make this an enum
            t.Param("SecurityLevel", t.uint8_t, "Security level of this data frame"),
            # ToDo: Make this an enum
            t.Param("KeyIdMode", t.uint8_t, "Key Id Mode of this frame"),
            t.Param("KeyIndex", t.uint8_t, "Key index of this frame"),
        ),
    )

    # send (on behalf of the next higher layer) a MAC disassociate confirm
    DisassociateCnf = t.CommandDef(
        t.CommandType.AREQ,
        0x87,
        rsp_schema=(
            t.Param("Status", t.Status, "Status is either Success (0) or Failure (1)"),
            t.Param(
                "DeviceAddrModeAddr",
                t.AddrModeAddress,
                "Address mode address of the device",
            ),
            t.Param("PanId", t.PanId, "The PAN Id of the device"),
        ),
    )

    # send (on behalf of the next higher layer) a MAC orphan indication
    OrphanInd = t.CommandDef(
        t.CommandType.AREQ,
        0x8A,
        rsp_schema=(
            t.Param("IEEE", t.EUI64, "Extended address of the orphan device"),
            t.Param("KeySource", t.KeySource, "Key Source of this data frame"),
            # ToDo: Make this an enum
            t.Param("SecurityLevel", t.uint8_t, "Security level of this data frame"),
            # ToDo: Make this an enum
            t.Param("KeyIdMode", t.uint8_t, "Key Id Mode of this frame"),
            t.Param("KeyIndex", t.uint8_t, "Key index of this frame"),
        ),
    )

    # send (on behalf of the next higher layer) a MAC poll confirmation
    PollCnf = t.CommandDef(t.CommandType.AREQ, 0x8B, rsp_schema=t.STATUS_SCHEMA)

    # TODO: investigate the actual structure of this command. The source code indicates
    #       that the response type differs heavily based on the ScanType.
    #       Also, ResultListMaxLength does not appear to be present.
    """
    # send (on behalf of the next higher layer) a MAC scan confirmation
    ScanCnf = t.CommandDef(
        t.CommandType.AREQ,
        0x8C,
        rsp_schema=(
            t.Param(
                "Status", t.Status, "Status is either Success (0) or Failure (1)"
            ),
            t.Param("ED", t.uint8_t, "ED max energy"),
            t.Param("ScanType", t.ScanType, "Specifies the scan type"),
            t.Param("ChannelPage", t.uint8_t, "The channel page to use"),
            t.Param(
                "UnScannedChannelList",
                t.Channels,
                "List of the un-scanned channels",
            ),
            t.Param(
                "ResultListCount", t.uint8_t, "Number of items in the result list"
            ),
            t.Param(
                "ResultListMaxLength",
                t.uint8_t,
                "Max length of the result list in bytes"
            ),
            t.Param("ResultList", t.LVList(t.uint8_t), "Result list"),
        ),
    )
    """

    # send (on behalf of the next higher layer) a MAC communication indicator
    CommStatusInd = t.CommandDef(
        t.CommandType.AREQ,
        0x8D,
        rsp_schema=(
            t.Param("Status", t.Status, "Status is either Success (0) or Failure (1)"),
            t.Param("DstAddrMode", t.AddrMode, "Destination address mode"),
            t.Param("SrcIEEE", t.EUI64, "Source address"),
            t.Param("DstIEEE", t.EUI64, "Destination address"),
            t.Param("TimeStamp", t.uint32_t, "The timestamp of the message"),
            t.Param("DevicePanId", t.PanId, "PAN Id of the device"),
            t.Param("Reason", t.uint8_t, "Reason of communication indication"),
            t.Param("KeySource", t.KeySource, "Key Source of this data frame"),
            # ToDo: Make this an enum
            t.Param("SecurityLevel", t.uint8_t, "Security level of this data frame"),
            # ToDo: Make this an enum
            t.Param("KeyIdMode", t.uint8_t, "Key Id Mode of this frame"),
            t.Param("KeyIndex", t.uint8_t, "Key index of this frame"),
        ),
    )

    # send (on behalf of the next higher layer) a MAC start confirmation
    StartCnf = t.CommandDef(t.CommandType.AREQ, 0x8E, rsp_schema=t.STATUS_SCHEMA)

    # send (on behalf of the next higher layer) a MAC Rx enable confirmation
    RxEnableCnf = t.CommandDef(t.CommandType.AREQ, 0x8F, rsp_schema=t.STATUS_SCHEMA)

    # send (on behalf of the next higher layer) a MAC purge confirmation
    PurgeCnf = t.CommandDef(
        t.CommandType.AREQ,
        0x9A,
        rsp_schema=(
            t.Param("Status", t.Status, "Status is either Success (0) or Failure (1)"),
            t.Param("Handle", t.uint8_t, "Handle of the message"),
        ),
    )



"""Serial bootloader interface. Implemented in sb_exec_v2.c
"""

import zigpy_znp.types as t

# Size of internal flash less 4 pages for boot loader,
# 6 pages for NV, & 1 page for lock bits.
IMAGE_SIZE = 0x40000 - 0x2000 - 0x3000 - 0x0800
IMAGE_CRC_OFFSET = 0x90

FLASH_WORD_SIZE = 4


class BootloaderStatus(t.enum_uint8):
    SUCCESS = 0
    FAILURE = 1
    INVALID_FCS = 2
    INVALID_FILE = 3
    FILESYSTEM_ERROR = 4
    ALREADY_STARTED = 5
    NO_RESPOSNE = 6
    VALIDATE_FAILED = 7
    CANCELED = 8


class BootloaderDeviceType(t.enum_uint8):
    CC2538 = 1
    CC2530 = 2


class BootloaderRunMode(t.enum_uint8):
    # Read the code, not the spec
    FORCE_BOOT = 0x10
    FORCE_RUN = FORCE_BOOT ^ 0xFF


class UBL(t.CommandsBase, subsystem=t.Subsystem.UBL_FUNC):
    WriteReq = t.CommandDef(
        t.CommandType.AREQ,
        0x01,
        req_schema=(
            (
                t.Param("FlashWordAddr", t.uint16_t, "Write address, in flash words"),
                t.Param(
                    "Data", t.TrailingBytes, "HandshakeRsp.BufferSize bytes of data"
                ),
            )
        ),
    )

    WriteRsp = t.CommandDef(
        t.CommandType.AREQ,
        0x81,
        rsp_schema=((t.Param("Status", BootloaderStatus, "Write status"),)),
    )

    ReadReq = t.CommandDef(
        t.CommandType.AREQ,
        0x02,
        req_schema=(
            t.Param("FlashWordAddr", t.uint16_t, "Read address, in flash words"),
        ),
    )

    ReadRsp = t.CommandDef(
        t.CommandType.AREQ,
        0x82,
        rsp_schema=(
            t.Param("Status", BootloaderStatus, "Read status"),
            # These are missing if the request is bad
            t.Param(
                "FlashWordAddr",
                t.uint16_t,
                "Address read from, in flash words",
                optional=True,
            ),
            t.Param(
                "Data",
                t.TrailingBytes,
                "HandshakeRsp.BufferSize bytes of data",
                optional=True,
            ),
        ),
    )

    EnableReq = t.CommandDef(t.CommandType.AREQ, 0x03, req_schema=())

    EnableRsp = t.CommandDef(
        t.CommandType.AREQ,
        0x83,
        rsp_schema=(t.Param("Status", BootloaderStatus, "Enable status"),),
    )

    HandshakeReq = t.CommandDef(t.CommandType.AREQ, 0x04, req_schema=())

    HandshakeRsp = t.CommandDef(
        t.CommandType.AREQ,
        0x84,
        rsp_schema=(
            t.Param("Status", BootloaderStatus, "Handshake status"),
            t.Param("BootloaderRevision", t.uint32_t, "Bootloader revision"),
            t.Param("DeviceType", BootloaderDeviceType, "Device type"),
            t.Param("BufferSize", t.uint32_t, "Read/write buffer size"),
            t.Param("PageSize", t.uint32_t, "Device page size"),
            t.Param("BootloaderCodeRevision", t.uint32_t, "Bootloader code revision"),
        ),
    )



import zigpy_znp.types as t
import zigpy_znp.commands as c
from zigpy_znp.types import nvids
from zigpy_znp.exceptions import SecurityError, InvalidCommandResponse

# Some NVIDs don't really exist and Z-Stack doesn't behave consistently when operations
# are performed on them.
PROXIED_NVIDS = {nvids.OsalNvIds.POLL_RATE_OLD16}


def serialize(value) -> bytes:
    if hasattr(value, "serialize"):
        value = value.serialize()
    elif not isinstance(value, (bytes, bytearray)):
        raise TypeError(
            f"Only bytes or serializable types can be written to NVRAM."
            f" Got {value!r} (type {type(value)})"
        )

    if not value:
        raise ValueError("NVRAM value cannot be empty")

    return value


class NVRAMHelper:
    def __init__(self, znp):
        self.znp = znp

    async def osal_delete(self, nv_id: t.uint16_t) -> bool:
        """
        Deletes an item from NVRAM. Returns whether or not the item existed.
        """

        length = (await self.znp.request(c.SYS.OSALNVLength.Req(Id=nv_id))).ItemLen

        if length == 0:
            return False

        delete_rsp = await self.znp.request(
            c.SYS.OSALNVDelete.Req(Id=nv_id, ItemLen=length)
        )

        return delete_rsp.Status == t.Status.SUCCESS

    async def osal_write(self, nv_id: t.uint16_t, value, *, create: bool = False):
        """
        Writes a complete value to NVRAM, optionally resizing and creating the item if
        necessary.

        Serializes all serializable values and passes bytes directly.
        """

        value = serialize(value)
        length = (await self.znp.request(c.SYS.OSALNVLength.Req(Id=nv_id))).ItemLen

        # Recreate the item if the length is not correct
        if length != len(value) and nv_id not in PROXIED_NVIDS:
            if not create:
                if length == 0:
                    raise KeyError(f"NV item does not exist: {nv_id!r}")
                else:
                    raise ValueError(
                        f"Stored length and actual length differ:"
                        f" {length} != {len(value)}"
                    )

            if length != 0:
                await self.znp.request(
                    c.SYS.OSALNVDelete.Req(Id=nv_id, ItemLen=length),
                    RspStatus=t.Status.SUCCESS,
                )

            await self.znp.request(
                c.SYS.OSALNVItemInit.Req(
                    Id=nv_id,
                    ItemLen=len(value),
                    Value=t.ShortBytes(value[:244]),
                ),
                RspStatus=t.Status.NV_ITEM_UNINIT,
            )

        # 244 bytes is the most you can fit in a single `SYS.OSALNVWriteExt` command
        for offset in range(0, len(value), 244):
            await self.znp.request(
                c.SYS.OSALNVWriteExt.Req(
                    Id=nv_id,
                    Offset=offset,
                    Value=t.ShortBytes(value[offset : offset + 244]),
                ),
                RspStatus=t.Status.SUCCESS,
            )

    async def osal_read(self, nv_id: t.uint16_t) -> bytes:
        """
        Reads a complete value from NVRAM.

        Raises an `KeyError` error if the NVID doesn't exist.
        """

        # XXX: Some NVIDs don't really exist and Z-Stack behaves strangely with them
        if nv_id in PROXIED_NVIDS:
            read_rsp = await self.znp.request(
                c.SYS.OSALNVRead.Req(Id=nv_id, Offset=0),
                RspStatus=t.Status.SUCCESS,
            )

            return read_rsp.Value

        # Every item has a length, even missing ones
        length = (await self.znp.request(c.SYS.OSALNVLength.Req(Id=nv_id))).ItemLen

        if length == 0:
            raise KeyError(f"NV item does not exist: {nv_id!r}")

        data = b""

        try:
            while len(data) < length:
                read_rsp = await self.znp.request(
                    c.SYS.OSALNVReadExt.Req(Id=nv_id, Offset=len(data)),
                    RspStatus=t.Status.SUCCESS,
                )

                data += read_rsp.Value
        except InvalidCommandResponse as e:
            # Only expected status code is INVALID_PARAMETER
            assert e.response.Status == t.Status.INVALID_PARAMETER

            # Not all items can be read out due to security policies, though this can
            # easily be bypassed for some. The SAPI "ConfigId" is only 8 bits which
            # means some nvids are not able to read this way.
            if not self.znp.capabilities & t.MTCapabilities.CAP_SAPI or nv_id > 0xFF:
                raise SecurityError(
                    f"NV item cannot be read due to security constraints: {nv_id!r}"
                )

            read_rsp = await self.znp.request(
                c.SAPI.ZBReadConfiguration.Req(ConfigId=nv_id),
                RspStatus=t.Status.SUCCESS,
                RspConfigId=nv_id,
            )

            data = read_rsp.Value

        assert len(data) == length

        return data

    async def delete(
        self, sys_id: t.uint8_t, item_id: t.uint16_t, sub_id: t.uint16_t
    ) -> bool:
        """
        Deletes a subitem from NVRAM. Returns whether or not the item existed.
        """

        delete_rsp = await self.znp.request(
            c.SYS.NVDelete.Req(SysId=sys_id, ItemId=item_id, SubId=sub_id)
        )

        return delete_rsp.Status == t.Status.SUCCESS

    async def write(
        self,
        sys_id: t.uint8_t,
        item_id: t.uint16_t,
        sub_id: t.uint16_t,
        value,
        *,
        create: bool = True,
    ):
        """
        Writes a value to NVRAM for the specified subsystem, item, and subitem.

        Calls to OSALNVWrite(sub_id=1) in newer Z-Stack releases are really calls to
        NVWrite(sys_id=ZSTACK, item_id=LEGACY, sub_id=1) in the background.
        """

        value = serialize(value)
        length = (
            await self.znp.request(
                c.SYS.NVLength.Req(SysId=sys_id, ItemId=item_id, SubId=sub_id)
            )
        ).Length

        if length != len(value) and not (
            sys_id == nvids.NvSysIds.ZSTACK
            and item_id in PROXIED_NVIDS
            and sub_id == 0x0000
        ):
            if not create:
                if length == 0:
                    raise KeyError(
                        f"NV item does not exist:"
                        f" sys_id={sys_id!r} item_id={item_id!r} sub_id={sub_id!r}"
                    )
                else:
                    raise ValueError(
                        f"Stored length and actual length differ:"
                        f" {length} != {len(value)}"
                    )

            if length != 0:
                await self.znp.request(
                    c.SYS.NVDelete.Req(SysId=sys_id, ItemId=item_id, SubId=sub_id),
                    RspStatus=t.Status.SUCCESS,
                )

            create_rsp = await self.znp.request(
                c.SYS.NVCreate.Req(
                    SysId=sys_id,
                    ItemId=item_id,
                    SubId=sub_id,
                    Length=len(value),
                )
            )

            if create_rsp.Status not in (t.Status.SUCCESS, t.Status.NV_ITEM_UNINIT):
                raise InvalidCommandResponse("Bad create status", create_rsp)

        # 244 bytes is the most you can fit in a single `SYS.NVWrite` command
        for offset in range(0, len(value), 244):
            await self.znp.request(
                c.SYS.NVWrite.Req(
                    SysId=sys_id,
                    ItemId=item_id,
                    SubId=sub_id,
                    Value=t.ShortBytes(value[offset : offset + 244]),
                    Offset=0,
                ),
                RspStatus=t.Status.SUCCESS,
            )

    async def read(
        self, sys_id: t.uint8_t, item_id: t.uint16_t, sub_id: t.uint16_t
    ) -> bytes:
        """
        Reads a value from NVRAM for the specified subsystem, item, and subitem.

        Calls to OSALNVRead(sub_id=1) in newer Z-Stack releases are really calls to
        NVRead(sys_id=ZSTACK, item_id=LEGACY, sub_id=1) in the background.

        Raises an `KeyError` error if the NVID doesn't exist.
        """

        length_rsp = await self.znp.request(
            c.SYS.NVLength.Req(SysId=sys_id, ItemId=item_id, SubId=sub_id)
        )
        length = length_rsp.Length

        if length == 0:
            raise KeyError(
                f"NV item does not exist:"
                f" sys_id={sys_id!r} item_id={item_id!r} sub_id={sub_id!r}"
            )

        data = b""

        while len(data) < length:
            read_rsp = await self.znp.request(
                c.SYS.NVRead.Req(
                    SysId=sys_id,
                    ItemId=item_id,
                    SubId=sub_id,
                    Offset=len(data),
                    Length=length,
                ),
                RspStatus=t.Status.SUCCESS,
            )

            data += read_rsp.Value

        assert len(data) == length

        return data



import typing

import zigpy_znp.types as t


class NwkState8(t.enum_uint8):
    NWK_INIT = 0
    NWK_JOINING_ORPHAN = 1
    NWK_DISC = 2
    NWK_JOINING = 3
    NWK_ENDDEVICE = 4
    PAN_CHNL_SELECTION = 5
    PAN_CHNL_VERIFY = 6
    PAN_STARTING = 7
    NWK_ROUTER = 8
    NWK_REJOINING = 9


class NwkState16(t.enum_uint16):
    """
    This enum is identical to `NwkState8`, with the only difference being the base class
    """

    NWK_INIT = 0
    NWK_JOINING_ORPHAN = 1
    NWK_DISC = 2
    NWK_JOINING = 3
    NWK_ENDDEVICE = 4
    PAN_CHNL_SELECTION = 5
    PAN_CHNL_VERIFY = 6
    PAN_STARTING = 7
    NWK_ROUTER = 8
    NWK_REJOINING = 9


class Empty(t.Bytes):
    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls, *args, **kwargs)

        if instance:
            raise ValueError("Empty must be empty")

        return instance

    @classmethod
    def deserialize(cls, data: bytes) -> typing.Tuple[t.Bytes, bytes]:
        return cls(), data


class NIB(t.Struct):
    SequenceNum: t.uint8_t
    PassiveAckTimeout: t.uint8_t
    MaxBroadcastRetries: t.uint8_t
    MaxChildren: t.uint8_t
    MaxDepth: t.uint8_t
    MaxRouters: t.uint8_t
    dummyNeighborTable: t.uint8_t
    BroadcastDeliveryTime: t.uint8_t
    ReportConstantCost: t.uint8_t
    RouteDiscRetries: t.uint8_t
    dummyRoutingTable: t.uint8_t
    SecureAllFrames: t.uint8_t
    SecurityLevel: t.uint8_t
    SymLink: t.uint8_t
    CapabilityFlags: t.uint8_t

    PaddingByte0: t.PaddingByte

    TransactionPersistenceTime: t.uint16_t

    nwkProtocolVersion: t.uint8_t
    RouteDiscoveryTime: t.uint8_t
    RouteExpiryTime: t.uint8_t

    PaddingByte1: t.PaddingByte

    nwkDevAddress: t.NWK

    nwkLogicalChannel: t.uint8_t

    PaddingByte2: t.PaddingByte

    nwkCoordAddress: t.NWK
    nwkCoordExtAddress: t.EUI64
    nwkPanId: t.uint16_t
    nwkState: NwkState16
    channelList: t.Channels

    beaconOrder: t.uint8_t
    superFrameOrder: t.uint8_t
    scanDuration: t.uint8_t
    battLifeExt: t.uint8_t

    allocatedRouterAddresses: t.uint32_t
    allocatedEndDeviceAddresses: t.uint32_t

    nodeDepth: t.uint8_t

    extendedPANID: t.EUI64

    nwkKeyLoaded: t.Bool

    spare1: t.NwkKeyDesc
    spare2: t.NwkKeyDesc

    spare3: t.uint8_t
    spare4: t.uint8_t

    nwkLinkStatusPeriod: t.uint8_t
    nwkRouterAgeLimit: t.uint8_t
    nwkUseMultiCast: t.Bool
    nwkIsConcentrator: t.Bool
    nwkConcentratorDiscoveryTime: t.uint8_t
    nwkConcentratorRadius: t.uint8_t
    nwkAllFresh: t.uint8_t

    PaddingByte3: t.PaddingByte

    nwkManagerAddr: t.NWK
    nwkTotalTransmissions: t.uint16_t
    nwkUpdateId: t.uint8_t

    PaddingByte4: t.PaddingByte


class CC2531NIB(t.Struct):
    """
    Struct doesn't allow field re-ordering so this is unfortunately a duplicate of the
    above `NIB` structure, with the following differences:

     1. There are no padding bytes.
     2. The `NwkState` enum is a `uint8_t` instead of a `uint16_t`

    Otherwise, the contents of the two structs are identical. These are only alignment
    differences between the two platforms.
    """

    SequenceNum: t.uint8_t
    PassiveAckTimeout: t.uint8_t
    MaxBroadcastRetries: t.uint8_t
    MaxChildren: t.uint8_t
    MaxDepth: t.uint8_t
    MaxRouters: t.uint8_t
    dummyNeighborTable: t.uint8_t
    BroadcastDeliveryTime: t.uint8_t
    ReportConstantCost: t.uint8_t
    RouteDiscRetries: t.uint8_t
    dummyRoutingTable: t.uint8_t
    SecureAllFrames: t.uint8_t
    SecurityLevel: t.uint8_t
    SymLink: t.uint8_t
    CapabilityFlags: t.uint8_t

    TransactionPersistenceTime: t.uint16_t

    nwkProtocolVersion: t.uint8_t
    RouteDiscoveryTime: t.uint8_t
    RouteExpiryTime: t.uint8_t

    nwkDevAddress: t.NWK

    nwkLogicalChannel: t.uint8_t

    nwkCoordAddress: t.NWK
    nwkCoordExtAddress: t.EUI64
    nwkPanId: t.uint16_t
    nwkState: NwkState8
    channelList: t.Channels

    beaconOrder: t.uint8_t
    superFrameOrder: t.uint8_t
    scanDuration: t.uint8_t
    battLifeExt: t.uint8_t

    allocatedRouterAddresses: t.uint32_t
    allocatedEndDeviceAddresses: t.uint32_t

    nodeDepth: t.uint8_t

    extendedPANID: t.EUI64

    nwkKeyLoaded: t.Bool

    spare1: t.NwkKeyDesc
    spare2: t.NwkKeyDesc

    spare3: t.uint8_t
    spare4: t.uint8_t

    nwkLinkStatusPeriod: t.uint8_t
    nwkRouterAgeLimit: t.uint8_t
    nwkUseMultiCast: t.Bool
    nwkIsConcentrator: t.Bool
    nwkConcentratorDiscoveryTime: t.uint8_t
    nwkConcentratorRadius: t.uint8_t
    nwkAllFresh: t.uint8_t

    nwkManagerAddr: t.NWK
    nwkTotalTransmissions: t.uint16_t
    nwkUpdateId: t.uint8_t


def parse_nib(data: bytes) -> typing.Union[NIB, CC2531NIB]:
    if len(data) == 116:
        nib, remaining = NIB.deserialize(data)
    elif len(data) == 110:
        nib, remaining = CC2531NIB.deserialize(data)
    else:
        raise ValueError(f"Unknown NIB format: {data!r}")

    assert not remaining

    return nib



from zigpy.zdo.types import ZDOCmd

import zigpy_znp.commands as c

"""
Zigpy expects to be able to directly send and receive ZDO commands.
Z-Stack, however, intercepts all responses and rewrites them to use its MT command set.
We use setup proxy functions that rewrite requests and responses based on their cluster.
"""


# The structure of this dict is:
#     {
#         cluster: (
#             zigpy_req_to_mt_req_converter,
#             mt_rsp_callback_matcher,
#             mt_rsp_callback_to_zigpy_rsp_converter,
#         )
#

ZDO_CONVERTERS = {
    ZDOCmd.Node_Desc_req: (
        (
            lambda addr, NWKAddrOfInterest: c.ZDO.NodeDescReq.Req(
                DstAddr=addr.address, NWKAddrOfInterest=NWKAddrOfInterest
            )
        ),
        (lambda addr: c.ZDO.NodeDescRsp.Callback(partial=True, Src=addr.address)),
        (
            lambda rsp: (
                ZDOCmd.Node_Desc_rsp,
                {
                    "Status": rsp.Status,
                    "NWKAddrOfInterest": rsp.NWK,
                    "NodeDescriptor": rsp.NodeDescriptor,
                },
            )
        ),
    ),
    ZDOCmd.Active_EP_req: (
        (
            lambda addr, NWKAddrOfInterest: c.ZDO.ActiveEpReq.Req(
                DstAddr=addr.address, NWKAddrOfInterest=NWKAddrOfInterest
            )
        ),
        (lambda addr: c.ZDO.ActiveEpRsp.Callback(partial=True, Src=addr.address)),
        (
            lambda rsp: (
                ZDOCmd.Active_EP_rsp,
                {
                    "Status": rsp.Status,
                    "NWKAddrOfInterest": rsp.NWK,
                    "ActiveEPList": rsp.ActiveEndpoints,
                },
            )
        ),
    ),
    ZDOCmd.Simple_Desc_req: (
        (
            lambda addr, NWKAddrOfInterest, EndPoint: (
                c.ZDO.SimpleDescReq.Req(
                    DstAddr=addr.address,
                    NWKAddrOfInterest=NWKAddrOfInterest,
                    Endpoint=EndPoint,
                )
            )
        ),
        (lambda addr: c.ZDO.SimpleDescRsp.Callback(partial=True, Src=addr.address)),
        (
            lambda rsp: (
                ZDOCmd.Simple_Desc_rsp,
                {
                    "Status": rsp.Status,
                    "NWKAddrOfInterest": rsp.NWK,
                    "SimpleDescriptor": rsp.SimpleDescriptor,
                },
            )
        ),
    ),
    ZDOCmd.Mgmt_Permit_Joining_req: (
        (
            lambda addr, PermitDuration, TC_Significant: (
                c.ZDO.MgmtPermitJoinReq.Req(
                    AddrMode=addr.mode,
                    Dst=addr.address,
                    Duration=PermitDuration,
                    TCSignificance=TC_Significant,
                )
            )
        ),
        (lambda addr: c.ZDO.MgmtPermitJoinRsp.Callback(partial=True, Src=addr.address)),
        (lambda rsp: (ZDOCmd.Mgmt_Permit_Joining_rsp, {"Status": rsp.Status})),
    ),
    ZDOCmd.Mgmt_Leave_req: (
        (
            lambda addr, DeviceAddress, Options: c.ZDO.MgmtLeaveReq.Req(
                DstAddr=addr.address,
                IEEE=DeviceAddress,
                RemoveChildren_Rejoin=c.zdo.LeaveOptions(Options),
            )
        ),
        (lambda addr: c.ZDO.MgmtLeaveRsp.Callback(partial=True, Src=addr.address)),
        (lambda rsp: (ZDOCmd.Mgmt_Leave_rsp, {"Status": rsp.Status})),
    ),
    ZDOCmd.Bind_req: (
        (
            lambda addr, SrcAddress, SrcEndpoint, ClusterID, DstAddress: (
                c.ZDO.BindReq.Req(
                    Dst=addr.address,
                    Src=SrcAddress,
                    SrcEndpoint=SrcEndpoint,
                    ClusterId=ClusterID,
                    Address=DstAddress,
                )
            )
        ),
        (lambda addr: c.ZDO.BindRsp.Callback(partial=True, Src=addr.address)),
        (lambda rsp: (ZDOCmd.Bind_rsp, {"Status": rsp.Status})),
    ),
    ZDOCmd.Unbind_req: (
        (
            lambda addr, SrcAddress, SrcEndpoint, ClusterID, DstAddress: (
                c.ZDO.UnBindReq.Req(
                    Dst=addr.address,
                    Src=SrcAddress,
                    SrcEndpoint=SrcEndpoint,
                    ClusterId=ClusterID,
                    Address=DstAddress,
                )
            )
        ),
        (lambda addr: c.ZDO.UnBindRsp.Callback(partial=True, Src=addr.address)),
        (lambda rsp: (ZDOCmd.Unbind_rsp, {"Status": rsp.Status})),
    ),
    ZDOCmd.Mgmt_Lqi_req: (
        (
            lambda addr, StartIndex: (
                c.ZDO.MgmtLqiReq.Req(
                    Dst=addr.address,
                    StartIndex=StartIndex,
                )
            )
        ),
        (lambda addr: c.ZDO.MgmtLqiRsp.Callback(partial=True, Src=addr.address)),
        (
            lambda rsp: (
                ZDOCmd.Mgmt_Lqi_rsp,
                {"Status": rsp.Status, "Neighbors": rsp.Neighbors},
            )
        ),
    ),
    ZDOCmd.Mgmt_Rtg_req: (
        (
            lambda addr, StartIndex: (
                c.ZDO.MgmtRtgReq.Req(
                    Dst=addr.address,
                    StartIndex=StartIndex,
                )
            )
        ),
        (lambda addr: c.ZDO.MgmtRtgRsp.Callback(partial=True, Src=addr.address)),
        (lambda rsp: (ZDOCmd.Mgmt_Rtg_rsp, {"Status": rsp.Status})),
    ),
}



import zigpy_znp.types as t


class TransmitOptions(t.enum_flag_uint8):
    NONE = 0

    # Will force the message to use Wildcard ProfileID
    WILDCARD_PROFILEID = 0x02

    # Will force APS to callback to preprocess before calling NWK layer
    APS_PREPROCESS = 0x04
    LIMIT_CONCENTRATOR = 0x08
    ACK_REQUEST = 0x10

    # Suppress Route Discovery for intermediate routes (route discovery preformed for
    # initiating device)
    SUPPRESS_ROUTE_DISC_NETWORK = 0x20
    ENABLE_SECURITY = 0x40
    SKIP_ROUTING = 0x80


class LatencyReq(t.enum_uint8):
    NoLatencyReqs = 0x00
    FastBeacons = 0x01
    SlowBeacons = 0x02


class AF(t.CommandsBase, subsystem=t.Subsystem.AF):
    # This command enables the tester to register an application's endpoint description
    Register = t.CommandDef(
        t.CommandType.SREQ,
        0x00,
        req_schema=(
            t.Param("Endpoint", t.uint8_t, "Endpoint Id of the device"),
            t.Param("ProfileId", t.uint16_t, "Application Profile ID"),
            t.Param("DeviceId", t.uint16_t, "Device Description ID"),
            t.Param("DeviceVersion", t.uint8_t, "Device version number"),
            t.Param(
                "LatencyReq",
                LatencyReq,
                (
                    "Specifies latency reqs: 0x00 - None, "
                    "0x01 -- fast beacons, "
                    "0x02 -- slow beacons"
                ),
            ),
            t.Param("InputClusters", t.ClusterIdList, "Input cluster list"),
            t.Param("OutputClusters", t.ClusterIdList, "Output cluster list"),
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # This command is used by the tester to build and send a message through AF layer
    DataRequest = t.CommandDef(
        t.CommandType.SREQ,
        0x01,
        req_schema=(
            t.Param("DstAddr", t.NWK, "Short address of the destination device"),
            t.Param("DstEndpoint", t.uint8_t, "Endpoint of the destination device"),
            t.Param("SrcEndpoint", t.uint8_t, "Endpoint of the source device"),
            t.Param("ClusterId", t.ClusterId, "Cluster ID"),
            t.Param("TSN", t.uint8_t, "Transaction Sequence Number"),
            t.Param(
                "Options",
                TransmitOptions,
                (
                    "Transmit options bitmask: bit 4 -- APS Ack, "
                    "bit 5 -- Route Discovery, "
                    "bit 6 -- APS security, "
                    "bit 7 -- Skip routing"
                ),
            ),
            t.Param(
                "Radius",
                t.uint8_t,
                "Specifies the number of hops allowed delivering the message",
            ),
            t.Param("Data", t.ShortBytes, "Data request"),
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # This extended form of the AF_DATA_REQUEST must be used to send an
    # inter-pan message
    DataRequestExt = t.CommandDef(
        t.CommandType.SREQ,
        0x02,
        req_schema=(
            t.Param(
                "DstAddrModeAddress",
                t.AddrModeAddress,
                "Destination address mode and address",
            ),
            t.Param("DstEndpoint", t.uint8_t, "Endpoint of the destination device"),
            t.Param(
                "DstPanId",
                t.PanId,
                (
                    "PanId of the destination device: 0x0000==Intra-Pan, "
                    "otherwise Inter-Pan"
                ),
            ),
            t.Param("SrcEndpoint", t.uint8_t, "Endpoint of the source device"),
            t.Param("ClusterId", t.ClusterId, "Cluster ID"),
            t.Param("TSN", t.uint8_t, "Transaction Sequence Number"),
            t.Param(
                "Options",
                TransmitOptions,
                (
                    "Transmit options bitmask: bit 4 -- APS Ack, "
                    "bit 5 -- Route Discovery, "
                    "bit 6 -- APS security, "
                    "bit 7 -- Skip routing"
                ),
            ),
            t.Param(
                "Radius",
                t.uint8_t,
                "Specifies the number of hops allowed delivering the message",
            ),
            t.Param("Data", t.LongBytes, "Data request"),
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # This command is used by the tester to build and send a message through AF layer
    # using source routing
    DataRequestSrcRtg = t.CommandDef(
        t.CommandType.SREQ,
        0x03,
        req_schema=(
            t.Param("DstAddr", t.NWK, "Short address of the destination device"),
            t.Param("DstEndpoint", t.uint8_t, "Endpoint of the destination device"),
            t.Param("SrcEndpoint", t.uint8_t, "Endpoint of the source device"),
            t.Param("ClusterId", t.ClusterId, "Cluster ID"),
            t.Param("TSN", t.uint8_t, "Transaction Sequence Number"),
            t.Param(
                "Options",
                TransmitOptions,
                (
                    "Transmit options bitmask: bit 4 -- APS Ack, "
                    "bit 5 -- Route Discovery, "
                    "bit 6 -- APS security, "
                    "bit 7 -- Skip routing"
                ),
            ),
            t.Param(
                "Radius",
                t.uint8_t,
                "Specifies the number of hops allowed delivering the message",
            ),
            t.Param("SourceRoute", t.NWKList, "Relay list"),
            t.Param("Data", t.ShortBytes, "Data request"),
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # XXX: UNDOCUMENTED
    Delete = t.CommandDef(
        t.CommandType.SREQ,
        0x04,
        req_schema=(t.Param("Endpoint", t.uint8_t, "Application Endpoint to delete"),),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # Inter-Pan control command and data
    InterPanCtl = t.CommandDef(
        t.CommandType.SREQ,
        0x10,
        req_schema=(
            t.Param(
                "Command",
                t.InterPanCommand,
                (
                    "0x00 InterPanClr Proxy call to StubAPS_SetIntraPanChannel() to"
                    " switch channel back to the NIB-specified channel. "
                    "0x01 InterPanSet Proxy call to StubAPS_SetInterPanChannel() "
                    "with the 1-byte channel specified. "
                    "0x02 InterPanReg If the 1-byte Endpoint specified by the data "
                    "argument is found by invoking afFindEndPointDesc(), then proxy"
                    " a call to StubAPS_RegisterApp() with the pointer to the "
                    "endPointDesc_t found (i.e. the Endpoint must already be "
                    "registered with AF)"
                ),
            ),
            t.Param("Data", t.Bytes, "Data"),
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # Huge AF data request data buffer store command and data
    DataStore = t.CommandDef(
        t.CommandType.SREQ,
        0x11,
        req_schema=(
            t.Param(
                "Index",
                t.uint16_t,
                (
                    "Specifies the index into the outgoing data request data buffer"
                    "to start the storing of this chunk of data"
                ),
            ),
            t.Param("Data", t.ShortBytes, "Data"),
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # Huge AF incoming message data buffer retrieve command
    DataRetrieve = t.CommandDef(
        t.CommandType.SREQ,
        0x12,
        req_schema=(
            t.Param("TimeStamp", t.uint32_t, "The timestamp of the message"),
            t.Param(
                "Index",
                t.uint16_t,
                (
                    "Specifies the index into the outgoing data request data buffer"
                    "to start the storing of this chunk of data"
                ),
            ),
            t.Param(
                "Length",
                t.uint8_t,
                (
                    "A length of zero is special and triggers the freeing of the "
                    "corresponding incoming message"
                ),
            ),
        ),
        rsp_schema=(
            t.Param("Status", t.Status, "Status is either Success (0) or Failure (1)"),
            t.Param("Data", t.ShortBytes, "Data"),
        ),
    )

    # proxy for afAPSF_ConfigSet()
    APSFConfigSet = t.CommandDef(
        t.CommandType.SREQ,
        0x13,
        req_schema=(
            t.Param("Endpoint", t.uint8_t, "Endpoint for which to set fragmentation"),
            t.Param(
                "FrameDelay",
                t.uint8_t,
                "APS Fragmentation inter-frame delay in milliseconds",
            ),
            t.Param("WindowSize", t.uint8_t, "APS Fragmentation window size"),
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # AF Callbacks
    # This command is sent by the device to the user after it receives a data request
    DataConfirm = t.CommandDef(
        t.CommandType.AREQ,
        0x80,
        rsp_schema=(
            t.Param("Status", t.Status, "Status is either Success (0) or Failure (1)"),
            t.Param("Endpoint", t.uint8_t, "Endpoint of the device"),
            t.Param("TSN", t.uint8_t, "Transaction Sequence Number"),
        ),
    )

    # This callback message is in response to incoming data to any of the registered
    # endpoints on this device
    IncomingMsg = t.CommandDef(
        t.CommandType.AREQ,
        0x81,
        rsp_schema=(
            t.Param("GroupId", t.GroupId, "The group ID of the device"),
            t.Param("ClusterId", t.ClusterId, "Cluster ID"),
            t.Param(
                "SrcAddr", t.NWK, "Short address of the device sending the message"
            ),
            t.Param("SrcEndpoint", t.uint8_t, "Endpoint of the source device"),
            t.Param("DstEndpoint", t.uint8_t, "Endpoint of the destination device"),
            t.Param(
                "WasBroadcast", t.Bool, "Was the incoming message broadcast or not"
            ),
            t.Param("LQI", t.uint8_t, "Link quality measured during reception"),
            t.Param("SecurityUse", t.Bool, "Is security in use or not"),
            t.Param("TimeStamp", t.uint32_t, "The timestamp of the message"),
            t.Param("TSN", t.uint8_t, "Transaction Sequence Number"),
            t.Param("Data", t.ShortBytes, "Data"),
            # https://e2e.ti.com/support/wireless-connectivity/zigbee-and-thread/f/158/t/455787
            t.Param("MacSrcAddr", t.NWK, "UNDOCUMENTED: MAC Source address"),
            t.Param(
                "MsgResultRadius", t.uint8_t, "UNDOCUMENTED: Messages result radius"
            ),
        ),
    )

    # This callback message is in response to incoming data to any of the registered
    # endpoints on this device when the code is compiled with the INTER_PAN
    # flag defined
    IncomingMsgExt = t.CommandDef(
        t.CommandType.AREQ,
        0x82,
        rsp_schema=(
            t.Param("GroupId", t.GroupId, "The group ID of the device"),
            t.Param("ClusterId", t.ClusterId, "Cluster ID"),
            t.Param(
                "SrcAddrModeAddress",
                t.AddrModeAddress,
                "Address of the device sending the message",
            ),
            t.Param("SrcEndpoint", t.uint8_t, "Endpoint of the source device"),
            t.Param("SrcPanId", t.PanId, "Source PanId of the message"),
            t.Param("DstEndpoint", t.uint8_t, "Endpoint of the destination device"),
            t.Param(
                "WasBroadcast", t.Bool, "Was the incoming message broadcast or not"
            ),
            t.Param("LQI", t.uint8_t, "Link quality measured during reception"),
            t.Param("SecurityUse", t.uint8_t, "Is security in use or not"),
            t.Param("TimeStamp", t.uint32_t, "The timestamp of the message"),
            t.Param("TSN", t.uint8_t, "Transaction Sequence Number"),
            t.Param("Data", t.LongBytes, "Data"),
        ),
    )

    # sent by the device to the user when it determines that an error occurred during
    # a reflected message
    ReflectError = t.CommandDef(
        t.CommandType.AREQ,
        0x83,
        rsp_schema=(
            t.Param("Status", t.Status, "Status is either Success (0) or Failure (1)"),
            t.Param("Endpoint", t.uint8_t, "Endpoint of the device"),
            t.Param("TSN", t.uint8_t, "Transaction Sequence Number"),
            t.Param("AddrMode", t.AddrMode, "Format of the address"),
            t.Param("Dst", t.NWK, "Destination address -- depends on AddrMode"),
        ),
    )



import enum
import typing


class Bytes(bytes):
    def serialize(self) -> "Bytes":
        return self

    @classmethod
    def deserialize(cls, data: bytes) -> typing.Tuple["Bytes", bytes]:
        return cls(data), b""

    def __repr__(self) -> str:
        # Reading byte sequences like \x200\x21 is extremely annoying
        # compared to \x20\x30\x21
        escaped = "".join(f"\\x{b:02X}" for b in self)

        return f"b'{escaped}'"

    __str__ = __repr__


class TrailingBytes(Bytes):
    """
    Bytes must occur at the very end of a parameter list for easy parsing.
    """

    pass


def serialize_list(objects) -> Bytes:
    return Bytes(b"".join([o.serialize() for o in objects]))


class FixedIntType(int):
    _signed = None
    _size = None

    def __new__(cls, *args, **kwargs):
        if cls._signed is None or cls._size is None:
            raise TypeError(f"{cls} is abstract and cannot be created")

        instance = super().__new__(cls, *args, **kwargs)
        instance.serialize()

        return instance

    def __init_subclass__(cls, signed=None, size=None, hex_repr=None) -> None:
        super().__init_subclass__()

        if signed is not None:
            cls._signed = signed

        if size is not None:
            cls._size = size

        if hex_repr:
            fmt = f"0x{{:0{cls._size * 2}X}}"
            cls.__str__ = cls.__repr__ = lambda self: fmt.format(self)
        elif hex_repr is not None and not hex_repr:
            cls.__str__ = super().__str__
            cls.__repr__ = super().__repr__

        # XXX: The enum module uses the first class with __new__ in its __dict__ as the
        #      member type. We have to ensure this is true for every subclass.
        if "__new__" not in cls.__dict__:
            cls.__new__ = cls.__new__

    def serialize(self) -> bytes:
        try:
            return self.to_bytes(self._size, "little", signed=self._signed)
        except OverflowError as e:
            # OverflowError is not a subclass of ValueError, making it annoying to catch
            raise ValueError(str(e)) from e

    @classmethod
    def deserialize(cls, data: bytes) -> typing.Tuple["FixedIntType", bytes]:
        if len(data) < cls._size:
            raise ValueError(f"Data is too short to contain {cls._size} bytes")

        r = cls.from_bytes(data[: cls._size], "little", signed=cls._signed)
        data = data[cls._size :]
        return r, data


class uint_t(FixedIntType, signed=False):
    pass


class int_t(FixedIntType, signed=True):
    pass


class int8s(int_t, size=1):
    pass


class int16s(int_t, size=2):
    pass


class int24s(int_t, size=3):
    pass


class int32s(int_t, size=4):
    pass


class int40s(int_t, size=5):
    pass


class int48s(int_t, size=6):
    pass


class int56s(int_t, size=7):
    pass


class int64s(int_t, size=8):
    pass


class uint8_t(uint_t, size=1):
    pass


class uint16_t(uint_t, size=2):
    pass


class uint24_t(uint_t, size=3):
    pass


class uint32_t(uint_t, size=4):
    pass


class uint40_t(uint_t, size=5):
    pass


class uint48_t(uint_t, size=6):
    pass


class uint56_t(uint_t, size=7):
    pass


class uint64_t(uint_t, size=8):
    pass


class ShortBytes(Bytes):
    _header = uint8_t

    def serialize(self) -> "Bytes":
        return self._header(len(self)).serialize() + self

    @classmethod
    def deserialize(cls, data: bytes) -> typing.Tuple[Bytes, bytes]:
        length, data = cls._header.deserialize(data)
        if length > len(data):
            raise ValueError(f"Data is too short to contain {length} bytes of data")
        return cls(data[:length]), data[length:]


class LongBytes(ShortBytes):
    _header = uint16_t


class LVList(list):
    _item_type = None
    _header = None

    def __init_subclass__(cls, *, item_type, length_type) -> None:
        super().__init_subclass__()
        cls._item_type = item_type
        cls._header = length_type

    def serialize(self) -> bytes:
        assert self._item_type is not None
        return self._header(len(self)).serialize() + serialize_list(
            [self._item_type(i) for i in self]
        )

    @classmethod
    def deserialize(cls, data: bytes) -> typing.Tuple["LVList", bytes]:
        assert cls._item_type is not None
        length, data = cls._header.deserialize(data)
        r = cls()
        for i in range(length):
            item, data = cls._item_type.deserialize(data)
            r.append(item)
        return r, data


class FixedList(list):
    _item_type = None
    _length = None

    def __init_subclass__(cls, *, item_type, length) -> None:
        super().__init_subclass__()
        cls._item_type = item_type
        cls._length = length

    def serialize(self) -> bytes:
        assert self._length is not None

        if len(self) != self._length:
            raise ValueError(
                f"Invalid length for {self!r}: expected {self._length}, got {len(self)}"
            )

        return serialize_list([self._item_type(i) for i in self])

    @classmethod
    def deserialize(cls, data: bytes) -> typing.Tuple["FixedList", bytes]:
        assert cls._item_type is not None
        r = cls()
        for i in range(cls._length):
            item, data = cls._item_type.deserialize(data)
            r.append(item)
        return r, data


def enum_flag_factory(int_type: FixedIntType) -> enum.Flag:
    """
    Mixins are broken by Python 3.8.6 so we must dynamically create the enum with the
    appropriate methods but with only one non-Enum parent class.
    """

    class _NewEnum(int_type, enum.Flag):
        # Rebind classmethods to our own class
        _missing_ = classmethod(enum.IntFlag._missing_.__func__)
        _create_pseudo_member_ = classmethod(
            enum.IntFlag._create_pseudo_member_.__func__
        )

        __or__ = enum.IntFlag.__or__
        __and__ = enum.IntFlag.__and__
        __xor__ = enum.IntFlag.__xor__
        __ror__ = enum.IntFlag.__ror__
        __rand__ = enum.IntFlag.__rand__
        __rxor__ = enum.IntFlag.__rxor__
        __invert__ = enum.IntFlag.__invert__

    return _NewEnum


class enum_uint8(uint8_t, enum.Enum):
    pass


class enum_uint16(uint16_t, enum.Enum):
    pass


class enum_uint24(uint24_t, enum.Enum):
    pass


class enum_uint32(uint32_t, enum.Enum):
    pass


class enum_uint40(uint40_t, enum.Enum):
    pass


class enum_uint48(uint48_t, enum.Enum):
    pass


class enum_uint56(uint56_t, enum.Enum):
    pass


class enum_uint64(uint64_t, enum.Enum):
    pass


class enum_flag_uint8(enum_flag_factory(uint8_t)):
    pass


class enum_flag_uint16(enum_flag_factory(uint16_t)):
    pass


class enum_flag_uint24(enum_flag_factory(uint24_t)):
    pass


class enum_flag_uint32(enum_flag_factory(uint32_t)):
    pass


class enum_flag_uint40(enum_flag_factory(uint40_t)):
    pass


class enum_flag_uint48(enum_flag_factory(uint48_t)):
    pass


class enum_flag_uint56(enum_flag_factory(uint56_t)):
    pass


class enum_flag_uint64(enum_flag_factory(uint64_t)):
    pass



import typing
import asyncio
import logging
import itertools
import contextlib
import dataclasses
from collections import Counter, defaultdict

import async_timeout

import zigpy_znp.types as t
import zigpy_znp.config as conf
import zigpy_znp.logger as log
import zigpy_znp.commands as c
from zigpy_znp import uart
from zigpy_znp.nvram import NVRAMHelper
from zigpy_znp.frames import GeneralFrame
from zigpy_znp.exceptions import CommandNotRecognized, InvalidCommandResponse
from zigpy_znp.types.nvids import ExNvIds, NvSysIds

LOGGER = logging.getLogger(__name__)
AFTER_CONNECT_DELAY = 1  # seconds
STARTUP_DELAY = 1  # seconds


def _deduplicate_commands(
    commands: typing.Iterable[t.CommandBase],
) -> typing.Tuple[t.CommandBase]:
    """
    Deduplicates an iterable of commands by folding more-specific commands into less-
    specific commands. Used to avoid triggering callbacks multiple times per packet.
    """

    # We essentially need to find the "maximal" commands, if you treat the relationship
    # between two commands as a partial order.
    maximal_commands = []

    # Command matching as a relation forms a partially ordered set.
    for command in commands:
        for index, other_command in enumerate(maximal_commands):
            if other_command.matches(command):
                # If the other command matches us, we are redundant
                break
            elif command.matches(other_command):
                # If we match another command, we replace it
                maximal_commands[index] = command
                break
            else:
                # Otherwise, we keep looking
                continue  # pragma: no cover
        else:
            # If we matched nothing and nothing matched us, we extend the list
            maximal_commands.append(command)

    # The start of each chain is the maximal element
    return tuple(maximal_commands)


async def detect_zstack_version(znp) -> float:
    """
    Feature detects the major version of Z-Stack running on the device.
    """

    # Z-Stack 1.2 does not have the AppConfig subsystem
    if not znp.capabilities & t.MTCapabilities.CAP_APP_CNF:
        return 1.2

    try:
        # Only Z-Stack 3.30+ has the new NVRAM system
        await znp.nvram.read(
            sys_id=NvSysIds.ZSTACK,
            item_id=ExNvIds.TCLK_TABLE,
            sub_id=0x0000,
        )
        return 3.30
    except KeyError:
        return 3.30
    except CommandNotRecognized:
        return 3.0


@dataclasses.dataclass(frozen=True)
class BaseResponseListener:
    matching_commands: typing.Tuple[t.CommandBase]

    def __post_init__(self):
        commands = _deduplicate_commands(self.matching_commands)

        if not commands:
            raise ValueError("Cannot create a listener without any matching commands")

        # We're frozen so __setattr__ is disallowed
        object.__setattr__(self, "matching_commands", commands)

    def matching_headers(self) -> typing.Set[t.CommandHeader]:
        """
        Returns the set of Z-Stack MT command headers for all the matching commands.
        """

        return {response.header for response in self.matching_commands}

    def resolve(self, response: t.CommandBase) -> bool:
        """
        Attempts to resolve the listener with a given response. Can be called with any
        command as an argument, including ones we don't match.
        """

        if not any(c.matches(response) for c in self.matching_commands):
            return False

        return self._resolve(response)

    def _resolve(self, response: t.CommandBase) -> bool:
        """
        Implemented by subclasses to handle matched commands.

        Return value indicates whether or not the listener has actually resolved,
        which can sometimes be unavoidable.
        """

        raise NotImplementedError()  # pragma: no cover

    def cancel(self):
        """
        Implement by subclasses to cancel the listener.

        Return value indicates whether or not the listener is cancelable.
        """

        raise NotImplementedError()  # pragma: no cover


@dataclasses.dataclass(frozen=True)
class OneShotResponseListener(BaseResponseListener):
    """
    A response listener that resolves a single future exactly once.
    """

    future: asyncio.Future = dataclasses.field(
        default_factory=lambda: asyncio.get_running_loop().create_future()
    )

    def _resolve(self, response: t.CommandBase) -> bool:
        if self.future.done():
            # This happens if the UART receives multiple packets during the same
            # event loop step and all of them match this listener. Our Future's
            # add_done_callback will not fire synchronously and thus the listener
            # is never properly removed. This isn't going to break anything.
            LOGGER.debug("Future already has a result set: %s", self.future)
            return False

        self.future.set_result(response)
        return True

    def cancel(self):
        if not self.future.done():
            self.future.cancel()

        return True


@dataclasses.dataclass(frozen=True)
class CallbackResponseListener(BaseResponseListener):
    """
    A response listener with a sync or async callback that is never resolved.
    """

    callback: typing.Callable[[t.CommandBase], typing.Any]

    def _resolve(self, response: t.CommandBase) -> bool:
        try:
            result = self.callback(response)

            # Run coroutines in the background
            if asyncio.iscoroutine(result):
                asyncio.create_task(result)
        except Exception:
            LOGGER.warning(
                "Caught an exception while executing callback", exc_info=True
            )

        # Callbacks are always resolved
        return True

    def cancel(self):
        # You can't cancel a callback
        return False


class ZNP:
    def __init__(self, config: conf.ConfigType):
        self._uart = None
        self._app = None
        self._config = config

        self._listeners = defaultdict(list)
        self._sync_request_lock = asyncio.Lock()

        self.capabilities = None
        self.version = None

        self.nvram = NVRAMHelper(self)

    def set_application(self, app):
        assert self._app is None
        self._app = app

    @property
    def _port_path(self) -> str:
        return self._config[conf.CONF_DEVICE][conf.CONF_DEVICE_PATH]

    async def connect(self, *, test_port=True) -> None:
        """
        Connects to the device specified by the "device" section of the config dict.

        The `test_port` kwarg allows port testing to be disabled, mainly to get into the
        bootloader.
        """

        # So we cannot connect twice
        assert self._uart is None

        try:
            self._uart = await uart.connect(self._config[conf.CONF_DEVICE], self)

            LOGGER.debug("Waiting %ss before sending anything", AFTER_CONNECT_DELAY)
            await asyncio.sleep(AFTER_CONNECT_DELAY)

            if self._config[conf.CONF_ZNP_CONFIG][conf.CONF_SKIP_BOOTLOADER]:
                LOGGER.debug("Sending bootloader skip byte")

                # XXX: Z-Stack locks up if other radios try probing it first.
                #      Writing the bootloader skip byte a bunch of times (at least 167)
                #      appears to reset it.
                skip = bytes([c.ubl.BootloaderRunMode.FORCE_RUN])
                self._uart._transport_write(skip * 256)

            # We have to disable all non-bootloader commands to enter the serial
            # bootloader upon connecting to the UART.
            if test_port:
                # Some Z-Stack 3 devices don't like you sending data immediately after
                # opening the serial port. A small delay helps, but they also sometimes
                # send a reset indication message when they're ready.
                LOGGER.debug(
                    "Waiting %ss or until a reset indication is received", STARTUP_DELAY
                )

                try:
                    async with async_timeout.timeout(STARTUP_DELAY):
                        await self.wait_for_response(
                            c.SYS.ResetInd.Callback(partial=True)
                        )
                except asyncio.TimeoutError:
                    pass

                LOGGER.debug("Testing connection to %s", self._port_path)

                # Make sure that our port works
                self.capabilities = (await self.request(c.SYS.Ping.Req())).Capabilities
                self.version = await detect_zstack_version(self)

                LOGGER.debug("Detected Z-Stack %s", self.version)
        except (Exception, asyncio.CancelledError):
            LOGGER.debug("Connection to %s failed, cleaning up", self._port_path)
            self.close()
            raise

        LOGGER.debug(
            "Connected to %s at %s baud",
            self._uart._transport.serial.name,
            self._uart._transport.serial.baudrate,
        )

    def connection_made(self) -> None:
        """
        Called by the UART object when a connection has been made.
        """
        pass

    def connection_lost(self, exc) -> None:
        """
        Called by the UART object to indicate that the port was closed. Propagates up
        to the `ControllerApplication` that owns this ZNP instance.
        """

        LOGGER.debug("We were disconnected from %s: %s", self._port_path, exc)

        if self._app is not None:
            self._app.connection_lost(exc)

    def close(self) -> None:
        """
        Cleans up resources, namely the listener queues.

        Calling this will reset ZNP to the same internal state as a fresh ZNP instance.
        """

        self._app = None

        for header, listeners in self._listeners.items():
            for listener in listeners:
                listener.cancel()

        self._listeners.clear()
        self.version = None
        self.capabilities = None

        if self._uart is not None:
            self._uart.close()
            self._uart = None

    def remove_listener(self, listener: BaseResponseListener) -> None:
        """
        Unbinds a listener from ZNP.

        Used by `wait_for_responses` to remove listeners for completed futures,
        regardless of their completion reason.
        """

        # If ZNP is closed while it's still running, `self._listeners` will be empty.
        if not self._listeners:
            return

        LOGGER.log(log.TRACE, "Removing listener %s", listener)

        for header in listener.matching_headers():
            try:
                self._listeners[header].remove(listener)
            except ValueError:
                pass

            if not self._listeners[header]:
                LOGGER.log(
                    log.TRACE, "Cleaning up empty listener list for header %s", header
                )
                del self._listeners[header]

        counts = Counter()

        for listener in itertools.chain.from_iterable(self._listeners.values()):
            counts[type(listener)] += 1

        LOGGER.log(
            log.TRACE,
            "There are %d callbacks and %d one-shot listeners remaining",
            counts[CallbackResponseListener],
            counts[OneShotResponseListener],
        )

    def frame_received(self, frame: GeneralFrame) -> bool:
        """
        Called when a frame has been received. Returns whether or not the frame was
        handled by any listener.

        XXX: Can be called multiple times in a single event loop step!
        """

        command_cls = c.COMMANDS_BY_ID[frame.header]
        command = command_cls.from_frame(frame)

        LOGGER.debug("Received command: %s", command)

        matched = False
        one_shot_matched = False

        for listener in self._listeners[command.header]:
            # XXX: A single response should *not* resolve multiple one-shot listeners!
            #      `future.add_done_callback` doesn't remove our listeners synchronously
            #      so doesn't prevent this from happening.
            if one_shot_matched and isinstance(listener, OneShotResponseListener):
                continue

            if not listener.resolve(command):
                LOGGER.log(log.TRACE, "%s does not match %s", command, listener)
                continue

            matched = True
            LOGGER.log(log.TRACE, "%s matches %s", command, listener)

            if isinstance(listener, OneShotResponseListener):
                one_shot_matched = True

        if not matched:
            self._unhandled_command(command)

        return matched

    def _unhandled_command(self, command: t.CommandBase):
        """
        Called when a command that is not handled by any listener is received.
        """

        LOGGER.warning("Received an unhandled command: %s", command)

    @contextlib.asynccontextmanager
    async def capture_responses(self, responses):
        """
        Captures all matched responses in a queue within the context manager.
        """

        queue = asyncio.Queue()
        listener = self.callback_for_responses(responses, queue.put_nowait)

        try:
            yield queue
        finally:
            self.remove_listener(listener)

    @contextlib.asynccontextmanager
    async def capture_responses_once(self, responses):
        """
        Captures all matched responses in a queue within the context manager.
        """

        future, listener = self.wait_for_responses(responses, context=True)

        try:
            yield future
        finally:
            self.remove_listener(listener)

    def callback_for_responses(self, responses, callback) -> CallbackResponseListener:
        """
        Creates a callback listener that matches any of the provided responses.

        Only exists for consistency with `wait_for_responses`, since callbacks can be
        executed more than once.
        """

        listener = CallbackResponseListener(responses, callback=callback)

        LOGGER.log(log.TRACE, "Creating callback %s", listener)

        for header in listener.matching_headers():
            self._listeners[header].append(listener)

        return listener

    def callback_for_response(
        self, response: t.CommandBase, callback
    ) -> CallbackResponseListener:
        """
        Creates a callback listener for a single response.
        """

        return self.callback_for_responses([response], callback)

    def wait_for_responses(self, responses, *, context=False) -> asyncio.Future:
        """
        Creates a one-shot listener that matches any *one* of the given responses.
        """

        listener = OneShotResponseListener(responses)

        LOGGER.log(log.TRACE, "Creating one-shot listener %s", listener)

        for header in listener.matching_headers():
            self._listeners[header].append(listener)

        # Remove the listener when the future is done, not only when it gets a result
        listener.future.add_done_callback(lambda _: self.remove_listener(listener))

        if context:
            return listener.future, listener
        else:
            return listener.future

    def wait_for_response(self, response: t.CommandBase) -> asyncio.Future:
        """
        Creates a one-shot listener for a single response.
        """

        return self.wait_for_responses([response])

    async def request(self, request: t.CommandBase, **response_params) -> t.CommandBase:
        """
        Sends a SREQ/AREQ request and returns its SRSP (only for SREQ), failing if any
        of the SRSP's parameters don't match `response_params`.
        """

        # Common mistake is to do `znp.request(c.SYS.Ping())`
        if type(request) is not request.Req:
            raise ValueError(f"Cannot send a command that isn't a request: {request!r}")

        # Construct a partial response out of the `Rsp*` kwargs if one is provded
        if request.Rsp:
            renamed_response_params = {}

            for param, value in response_params.items():
                if not param.startswith("Rsp"):
                    raise KeyError(
                        f"All response params must start with 'Rsp': {param!r}"
                    )

                renamed_response_params[param.replace("Rsp", "", 1)] = value

            # Construct our response before we send the request so that we fail early
            partial_response = request.Rsp(partial=True, **renamed_response_params)
        elif response_params:
            raise ValueError(
                f"Command has no response so response_params={response_params} "
                f"will have no effect"
            )

        # We should only be sending one SREQ at a time, according to the spec
        async with self._sync_request_lock:
            LOGGER.debug("Sending request: %s", request)

            # If our request has no response, we cannot wait for one
            if not request.Rsp:
                LOGGER.debug("Request has no response, not waiting for one.")
                self._uart.send(request.to_frame())
                return

            # We need to create the response listener before we send the request
            response_future = self.wait_for_responses(
                [
                    request.Rsp(partial=True),
                    c.RPCError.CommandNotRecognized.Rsp(
                        partial=True, RequestHeader=request.header
                    ),
                ]
            )

            self._uart.send(request.to_frame())

            # We should get a SRSP in a reasonable amount of time
            async with async_timeout.timeout(
                self._config[conf.CONF_ZNP_CONFIG][conf.CONF_SREQ_TIMEOUT]
            ):
                # We lock until either a sync response is seen or an error occurs
                response = await response_future

        if isinstance(response, c.RPCError.CommandNotRecognized.Rsp):
            raise CommandNotRecognized(
                f"Fatal request error {response} in response to {request}"
            )

        # If the sync response we got is not what we wanted, this is an error
        if not partial_response.matches(response):
            raise InvalidCommandResponse(
                f"Expected SRSP response {partial_response}, got {response}", response
            )

        return response

    async def request_callback_rsp(
        self, *, request, callback, timeout=None, **response_params
    ):
        """
        Sends an SREQ, gets its SRSP confirmation, and waits for its real AREQ response.
        A bug-free version of:

            req_rsp = await req
            callback_rsp = await req_callback

        This is necessary because the SRSP and the AREQ may arrive in the same "chunk"
        from the UART and be handled in the same event loop step by ZNP.
        """

        if timeout is None:
            timeout = self._config[conf.CONF_ZNP_CONFIG][conf.CONF_ARSP_TIMEOUT]

        # The async context manager allows us to clean up resources upon cancellation
        async with self.capture_responses_once([callback]) as callback_rsp:
            await self.request(request, **response_params)

            async with async_timeout.timeout(timeout):
                return await callback_rsp



import sys
import json
import asyncio
import logging
import argparse

import zigpy_znp.types as t
import zigpy_znp.commands as c
from zigpy_znp.api import ZNP
from zigpy_znp.config import CONFIG_SCHEMA
from zigpy_znp.types.nvids import ExNvIds, NvSysIds, OsalNvIds
from zigpy_znp.tools.common import setup_parser

LOGGER = logging.getLogger(__name__)


async def restore(radio_path, backup):
    znp = ZNP(CONFIG_SCHEMA({"device": {"path": radio_path}}))

    await znp.connect()

    # First write the NVRAM items common to all radios
    for nwk_nvid, value in backup["LEGACY"].items():
        if "+" in nwk_nvid:
            nwk_nvid, _, offset = nwk_nvid.partition("+")
            offset = int(offset)
            nvid = OsalNvIds[nwk_nvid] + offset
        else:
            nvid = OsalNvIds[nwk_nvid]

        value = bytes.fromhex(value)
        await znp.nvram.osal_write(nvid, value, create=True)

    for item_name, sub_ids in backup.items():
        item_id = ExNvIds[item_name]

        if item_id == ExNvIds.LEGACY:
            continue

        for sub_id, value in sub_ids.items():
            sub_id = int(sub_id, 16)
            value = bytes.fromhex(value)

            await znp.nvram.write(
                sys_id=NvSysIds.ZSTACK,
                item_id=item_id,
                sub_id=sub_id,
                value=value,
                create=True,
            )

    # Reset afterwards to have the new values take effect
    await znp.request_callback_rsp(
        request=c.SYS.ResetReq.Req(Type=t.ResetType.Soft),
        callback=c.SYS.ResetInd.Callback(partial=True),
    )


async def main(argv):
    parser = setup_parser("Restore a radio's NVRAM from a previous backup")
    parser.add_argument(
        "--input", "-i", type=argparse.FileType("r"), help="Input file", required=True
    )

    args = parser.parse_args(argv)
    backup = json.load(args.input)
    await restore(args.serial, backup)


if __name__ == "__main__":
    asyncio.run(main(sys.argv[1:]))  # pragma: no cover



import zigpy_znp.types as t


class DiscreteCommand(t.enum_flag_uint8):
    ZDOStart = 0x40
    ResetNwk = 0x80


class ZNP(t.CommandsBase, subsystem=t.Subsystem.ZNP):
    BasicCfg = t.CommandDef(
        t.CommandType.SREQ,
        0x00,
        req_schema=(
            t.Param(
                "BasicRspRate",
                t.uint32_t,
                "Rate at which to generate the basic response",
            ),
            t.Param("ChannelList", t.Channels, "CHANLIST NV item"),
            t.Param("PanId", t.NWK, "PANID NV item"),
            t.Param("LogicalType", t.DeviceLogicalType, "LOGICAL_TYPE NV item"),
            t.Param("CmdDisc", DiscreteCommand, "Discrete command bits"),
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # These exist in the codebase but the functions are completely empty
    # ZclCfg = t.CommandDef(t.CommandType.SREQ, 0x10, req_schema=(), rsp_schema=())
    # SeCfg = t.CommandDef(t.CommandType.SREQ, 0x20, req_schema=(), rsp_schema=())



import typing
import inspect
import dataclasses

import zigpy_znp.types.basic as t

NoneType = type(None)


class ListSubclass(list):
    # So we can call `setattr()` on it
    pass


class Struct:
    @classmethod
    def real_cls(cls) -> type:
        # The "Optional" subclass is dynamically created and breaks types.
        # We have to use a little introspection to find our real class.
        return next(c for c in cls.__mro__ if c.__name__ != "Optional")

    @classmethod
    def _annotations(cls) -> typing.Dict[str, typing.Any]:
        # First get our proper subclasses
        subclasses = []

        for subcls in cls.real_cls().__mro__:
            if subcls is Struct:
                break

            subclasses.append(subcls)

        annotations = {}

        # Iterate over the annotations *backwards*.
        # We want subclasses' annotations to override their parent classes'.
        for subcls in subclasses[::-1]:
            annotations.update(getattr(subcls, "__annotations__", {}))

        return annotations

    def __init_subclass__(cls):
        super().__init_subclass__()

        # Explicitly check for old-style structs and fail very early
        if hasattr(cls, "_fields"):
            raise TypeError(
                "Struct subclasses do not use `_fields` anymore."
                " Use class attributes with type annotations."
            )

        # We generate fields up here to fail early (and cache it)
        real_cls = cls.real_cls()
        fields = real_cls.fields()

        # We dynamically create our subclass's `__new__` method
        def __new__(cls, *args, **kwargs) -> "Struct":
            # Like a copy constructor
            if len(args) == 1 and isinstance(args[0], real_cls):
                if kwargs:
                    raise ValueError(
                        f"Cannot use copy constructor with kwargs: " f"{kwargs!r}"
                    )

                kwargs = args[0].as_dict()
                args = ()

            # Pretend our signature is `__new__(cls, p1: t1, p2: t2, ...)`
            signature = inspect.Signature(
                parameters=[
                    inspect.Parameter(
                        name=f.name,
                        kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                        default=None,
                        annotation=f.type,
                    )
                    for f in real_cls.fields()
                ]
            )

            bound = signature.bind(*args, **kwargs)
            bound.apply_defaults()

            instance = super().__new__(real_cls)

            # Set and convert the attributes to their respective types
            for name, value in bound.arguments.items():
                field = getattr(fields, name)

                if value is not None:
                    field_type = field.get_type_for(instance)

                    try:
                        value = field_type(value)
                    except Exception as e:
                        raise ValueError(
                            f"Failed to convert {name}={value!r} from type"
                            f" {type(value)} to {field_type}"
                        ) from e

                setattr(instance, name, value)

            return instance

        # Finally, attach the above __new__ classmethod to our subclass
        cls.__new__ = __new__

    @classmethod
    def fields(cls) -> typing.List["StructField"]:
        fields = ListSubclass()
        seen_optional = False

        # We need both to throw type errors in case a field is not annotated
        annotations = cls.real_cls()._annotations()
        variables = vars(cls.real_cls())

        # `set(annotations) | set(variables)` doesn't preserve order, which we need
        for name in list(annotations) + [v for v in variables if v not in annotations]:
            # It's a lot easier to debug when things break immediately instead of
            # fields being silently skipped
            if hasattr(cls, name):
                field = getattr(cls, name)

                if not isinstance(field, StructField):
                    if name.startswith("_") or name.isupper():
                        # _foo and FOO are considered constants and ignored
                        continue
                    elif isinstance(field, property):
                        # Ignore properties
                        continue
                    elif inspect.isfunction(field) or inspect.ismethod(field):
                        # Ignore methods and overridden functions
                        continue

                    # Everything else is an error
                    raise TypeError(
                        f"Field {name!r}={field!r} is not a constant or a field"
                    )
            else:
                field = StructField(name)

            field = field.replace(name=name)

            if name in annotations:
                annotation = annotations[name]

                if field.type is not None and field.type != annotation:
                    raise TypeError(
                        f"Field {name!r} type annotation conflicts with provided type:"
                        f" {annotation} != {field.type}"
                    )

                field = field.replace(type=annotation)

            if field.optional:
                seen_optional = True

            if seen_optional and not field.optional:
                raise TypeError(
                    f"No required fields can come after optional fields: {field!r}"
                )

            fields.append(field)
            setattr(fields, field.name, field)

        return fields

    def assigned_fields(self, *, strict=False) -> typing.List["StructField"]:
        assigned_fields = ListSubclass()

        for field in self.fields():
            value = getattr(self, field.name)

            # Ignore fields that aren't required
            if field.requires is not None and not field.requires(self):
                continue

            # Missing non-optional required fields cause an error if strict
            if value is None:
                if field.optional or not strict:
                    continue

                raise ValueError(f"Value for field {field.name} is required")

            assigned_fields.append((field, value))
            setattr(assigned_fields, field.name, (field, value))

        return assigned_fields

    def as_dict(self) -> typing.Dict[str, typing.Any]:
        return {f.name: v for f, v in self.assigned_fields()}

    def serialize(self) -> bytes:
        return b"".join(
            f.get_type_for(self)(v).serialize()
            for f, v in self.assigned_fields(strict=True)
        )

    @classmethod
    def deserialize(cls, data: bytes) -> typing.Tuple["Struct", bytes]:
        instance = cls()

        for field in cls.fields():
            if field.requires is not None and not field.requires(instance):
                continue

            try:
                value, data = field.get_type_for(instance).deserialize(data)
            except (ValueError, AssertionError):
                if field.optional:
                    break

                raise

            setattr(instance, field.name, value)

        return instance, data

    def replace(self, **kwargs) -> "Struct":
        d = self.as_dict().copy()
        d.update(kwargs)

        return type(self)(**d)

    def __eq__(self, other: "Struct") -> bool:
        if not isinstance(self, type(other)) and not isinstance(other, type(self)):
            return False

        return self.as_dict() == other.as_dict()

    def __repr__(self) -> str:
        kwargs = ", ".join([f"{k}={v!r}" for k, v in self.as_dict().items()])
        return f"{type(self).__name__}({kwargs})"


@dataclasses.dataclass(frozen=True)
class StructField:
    name: typing.Optional[str] = None
    type: typing.Optional[type] = None

    dynamic_type: typing.Optional[typing.Callable[[Struct], type]] = None
    requires: typing.Optional[typing.Callable[[Struct], bool]] = None

    def __post_init__(self) -> None:
        # Fail to initialize if the concrete type is invalid
        if self.dynamic_type is None:
            self.concrete_type

    def get_type_for(self, struct: Struct) -> type:
        if self.dynamic_type is not None:
            try:
                return self.dynamic_type(struct)
            except ValueError:
                raise
            except Exception as e:
                raise ValueError(
                    f"Failed to get a type for field {self!r} from {struct!r}"
                ) from e

        return self.concrete_type

    @property
    def optional(self) -> bool:
        # typing.Optional[Foo] is really typing.Union[Foo, None]
        if getattr(self.type, "__origin__", None) is not typing.Union:
            return False

        # I can't think of a case where this is ever False but it's best to be explicit
        return NoneType in self.type.__args__

    @property
    def concrete_type(self) -> type:
        if getattr(self.type, "__origin__", None) is not typing.Union:
            return self.type

        types = set(self.type.__args__) - {NoneType}

        if len(types) > 1:
            raise TypeError(
                f"Struct field cannot have more than one concrete type: {types}"
            )

        return tuple(types)[0]

    def replace(self, **kwargs) -> "StructField":
        return dataclasses.replace(self, **kwargs)


class PaddingByte(t.Bytes):
    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls, *args, **kwargs)

        if len(instance) != 1:
            raise ValueError("Padding byte must be a single byte")

        return instance

    @classmethod
    def deserialize(cls, data: bytes) -> typing.Tuple[t.Bytes, bytes]:
        if not data:
            raise ValueError("Data is empty and cannot contain a padding byte")

        return cls(data[:1]), data[1:]



import logging
import argparse

import coloredlogs

import zigpy_znp.logger as log

LOG_LEVELS = [logging.INFO, logging.DEBUG, log._TRACE]


class CustomArgumentParser(argparse.ArgumentParser):
    def parse_args(self, args=None, namespace=None):
        args = super().parse_args(args, namespace)

        # Since we're running as a CLI tool, install our own log level and color logger
        log.TRACE = log._TRACE
        logging.addLevelName(log.TRACE, "TRACE")

        # But still allow the user to configure verbosity
        verbosity = args.verbosity
        log_level = LOG_LEVELS[min(max(0, verbosity), len(LOG_LEVELS) - 1)]

        # coloredlogs uses "spam" for level 5, not "trace"
        level_styles = coloredlogs.DEFAULT_LEVEL_STYLES.copy()
        level_styles["trace"] = level_styles["spam"]

        logging.getLogger().setLevel(log_level)

        coloredlogs.install(
            level=log_level,
            level_styles=level_styles,
        )

        return args


def setup_parser(description: str) -> argparse.ArgumentParser:
    """
    Creates an ArgumentParser that sets up a logger with a configurable verbosity
    and a positional serial port argument.
    """

    parser = CustomArgumentParser(
        description=description,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbosity",
        action="count",
        default=0,
        help="increases verbosity",
    )
    parser.add_argument("serial", type=str, help="Serial port path")

    return parser
